import struct
import numpy as np
from enum import IntEnum
from typing import BinaryIO, Optional
from vstools import get_render_progress, vs

__all__ = ["audio_async_render", "WaveHeader"]

"""
Most if not everything in this file is stolen from this because I don't feel like waiting for setsu to add it to vapoursynth itself.

https://github.com/Ichunjo/vardautomation/blob/fae054956b3611e641276dc92f4a8c4060a3d8e2/vardautomation/render.py
"""


class WaveFormat(IntEnum):
    """
    WAVE form wFormatTag IDs
    Complete list is in mmreg.h in Windows 10 SDK.
    """

    PCM = 0x0001
    IEEE_FLOAT = 0x0003
    EXTENSIBLE = 0xFFFE


class WaveHeader(IntEnum):
    """
    Wave headers
    """

    WAVE = 0
    WAVE64 = 1
    AUTO = 2


WAVE_RIFF_TAG = b"RIFF"
WAVE_WAVE_TAG = b"WAVE"
WAVE_FMT_TAG = b"fmt "
WAVE_DATA_TAG = b"data"

WAVE64_RIFF_UUID = (0x72, 0x69, 0x66, 0x66, 0x2E, 0x91, 0xCF, 0x11, 0xA5, 0xD6, 0x28, 0xDB, 0x04, 0xC1, 0x00, 0x00)
WAVE64_WAVE_UUID = (0x77, 0x61, 0x76, 0x65, 0xF3, 0xAC, 0xD3, 0x11, 0x8C, 0xD1, 0x00, 0xC0, 0x4F, 0x8E, 0xDB, 0x8A)
WAVE64_FMT_UUID = (0x66, 0x6D, 0x74, 0x20, 0xF3, 0xAC, 0xD3, 0x11, 0x8C, 0xD1, 0x00, 0xC0, 0x4F, 0x8E, 0xDB, 0x8A)
WAVE64_DATA_UUID = (0x64, 0x61, 0x74, 0x61, 0xF3, 0xAC, 0xD3, 0x11, 0x8C, 0xD1, 0x00, 0xC0, 0x4F, 0x8E, 0xDB, 0x8A)
WAVE_FMT_EXTENSIBLE_SUBFORMAT = (
    (WaveFormat.PCM, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71),
    (WaveFormat.IEEE_FLOAT, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71),
)


def audio_async_render(
    audio: vs.AudioNode, outfile: BinaryIO, header: WaveHeader = WaveHeader.AUTO, progress: Optional[str] = "Rendering audio..."
) -> None:
    """
    Render an audio by requesting frames asynchronously using audio.frames.

    Implementation-like of VideoNode.output for an AudioNode that isn't in the Cython side yet.

    :param audio:       Audio to render.
    :param outfile:     Render output BinaryIO handle.
    :param header:      Kind of Wave header.
                        WaveHeader.AUTO adds a Wave64 header if the audio

                        * Has more than 2 channels
                        * Has a bitdepth > 16
                        * Has more than 44100 samples

    :param progress:    String to use for render progress display.
                        If empty or ``None``, no progress display.
    """
    if progress:
        p = get_render_progress()
        task = p.add_task(progress, total=audio.num_frames)
        p.start()

    bytes_per_output_sample = (audio.bits_per_sample + 7) // 8
    block_align = audio.num_channels * bytes_per_output_sample
    bytes_per_second = audio.sample_rate * block_align
    data_size = audio.num_samples * block_align

    if header == WaveHeader.AUTO:
        conditions = (audio.num_channels > 2, audio.bits_per_sample > 16, audio.num_samples > 44100)
        header_func, use_w64 = (_w64_header, WaveHeader.WAVE64) if any(conditions) else (_wav_header, WaveHeader.WAVE)
    else:
        use_w64 = header
        header_func = (_wav_header, _w64_header)[header]

    outfile.write(header_func(audio, bytes_per_second, block_align, data_size))

    for f in audio.frames(close=True):
        if progress:
            p.update(task, advance=1)
        _finish_frame_audio(f, outfile, audio.bits_per_sample == 24)
    size = outfile.tell()
    if use_w64:
        outfile.seek(16)
        outfile.write(struct.pack("<Q", size))
    else:
        outfile.seek(4)
        outfile.write(struct.pack("<I", size - 8))
    if progress:
        p.stop()


def _wav_header(audio: vs.AudioNode, bps: int, block_align: int, data_size: int) -> bytes:
    header = WAVE_RIFF_TAG
    header += b"\x00\x00\x00\x00"
    header += WAVE_WAVE_TAG

    header += WAVE_FMT_TAG
    format_tag = WaveFormat.IEEE_FLOAT if audio.sample_type == vs.FLOAT else WaveFormat.PCM

    fmt_chunk_data = struct.pack("<HHIIHH", format_tag, audio.num_channels, audio.sample_rate, bps, block_align, audio.bits_per_sample)
    header += struct.pack("<I", len(fmt_chunk_data))
    header += fmt_chunk_data

    if len(header) + data_size > 0xFFFFFFFE:
        raise ValueError("Data exceeds wave file size limit")

    header += WAVE_DATA_TAG
    header += struct.pack("<I", data_size)
    return header


def _w64_header(audio: vs.AudioNode, bps: int, block_align: int, data_size: int) -> bytes:
    header = bytes(WAVE64_RIFF_UUID)
    header += b"\x00\x00\x00\x00\x00\x00\x00\x00"
    header += bytes(WAVE64_WAVE_UUID)
    fmt_guid = bytes(WAVE64_FMT_UUID)
    header += fmt_guid

    format_tag = WaveFormat.EXTENSIBLE

    cb_size = 22
    fmt_chunk_data = struct.pack(
        "<HHIIHHHHI",
        format_tag,
        audio.num_channels,
        audio.sample_rate,
        bps,
        block_align,
        audio.bits_per_sample,
        cb_size,
        audio.bits_per_sample,
        audio.channel_layout,
    )
    fmt_chunk_data += bytes(WAVE_FMT_EXTENSIBLE_SUBFORMAT[audio.sample_type])

    header += struct.pack("<Q", len(fmt_guid) + 8 + len(fmt_chunk_data))
    header += fmt_chunk_data

    data_uuid = bytes(WAVE64_DATA_UUID)
    header += data_uuid
    header += struct.pack("<Q", data_size + len(data_uuid) + 8)
    return header


def _finish_frame_audio(frame: vs.AudioFrame, outfile: BinaryIO, _24bit: bool) -> None:
    data = np.stack([frame[i] for i in range(frame.num_channels)], axis=1)

    if _24bit:
        if data.ndim == 1:
            data.shape += (1,)
        data = (data // 2**8).reshape(data.shape + (1,)) >> np.array([0, 8, 16], np.uint32)
        outfile.write(data.ravel().astype(np.uint8).tobytes())
    else:
        outfile.write(data.ravel().view(np.int8).tobytes())



from genericpath import isfile
import re
import os
from tracemalloc import start
from muxtools import error, PathLike, ensure_path, warn
from vstools import vs, Matrix, Primaries, Transfer, ColorRange, ChromaLocation, get_prop, Colorspace
from ..utils.types import Zone

__all__ = ["settings_builder_x265", "settings_builder_x264", "sb", "sb265", "sb264"]


def is_full_zone(zone: Zone) -> bool:
    if isinstance(zone[2], str):
        if len(zone) < 4:
            raise error(f"Zone '{zone}' is invalid.")
        return True
    else:
        return False


def shift_zones(zones: Zone | list[Zone] | None, start_frame: int = 0) -> list[Zone] | None:
    if not zones:
        return None
    if not isinstance(zones, list):
        zones = [zones]

    newzones = list[Zone]()

    for zone in zones:
        zone = list(zone)
        zone[0] = zone[0] - start_frame
        zone[1] = zone[1] - start_frame
        if zone[1] < 0:
            continue
        if zone[0] < 0:
            zone[0] = 0
        zone = tuple(zone)
        newzones.append(zone)

    return newzones


def zones_to_args(zones: Zone | list[Zone] | None, x265: bool) -> list[str]:
    args: list[str] = []
    if not zones:
        return args
    if not isinstance(zones, list):
        zones = [zones]
    zones_settings: str = ""
    for i, zone in enumerate(zones):
        if is_full_zone(zone):
            if x265 and zone[2].lower() not in ["q", "b"]:
                raise error(f"Zone '{zone}' is invalid for x265. Please only use b or q.")
            zones_settings += f"{zone[0]},{zone[1]},{zone[2]}={zone[3]}"
        else:
            zones_settings += f"{zone[0]},{zone[1]},b={zone[2]}"
        if i != len(zones) - 1:
            zones_settings += "/"
    args.extend(["--zones", zones_settings])
    return args


def settings_builder_x265(
    preset: str | int = "slower",
    crf: float = 14.0,
    qcomp: float = 0.75,
    psy_rd: float = 2.0,
    psy_rdoq: float = 2.0,
    aq_strength: float = 0.75,
    aq_mode: int = 3,
    rd: int = 4,
    rect: bool = True,
    amp: bool = False,
    chroma_qpoffsets: int = -2,
    tu_intra_depth: int = 2,
    tu_inter_depth: int = 2,
    rskip: bool | int = 0,
    tskip: bool = False,
    ref: int = 4,
    bframes: int = 16,
    cutree: bool = False,
    rc_lookahead: int = 60,
    subme: int = 5,
    me: int = 3,
    b_intra: bool = True,
    weightb: bool = True,
    deblock: list[int] | str = [-2, -2],
    append: str = "",
    **kwargs,
) -> str:
    # Simple insert values
    settings = f" --preset {preset} --crf {crf} --bframes {bframes} --ref {ref} --rc-lookahead {rc_lookahead} --subme {subme} --me {me}"
    settings += f" --aq-mode {aq_mode} --aq-strength {aq_strength} --qcomp {qcomp} --cbqpoffs {chroma_qpoffsets} --crqpoffs {chroma_qpoffsets}"
    settings += f" --rd {rd} --psy-rd {psy_rd} --psy-rdoq {psy_rdoq} --tu-intra-depth {tu_intra_depth} --tu-inter-depth {tu_inter_depth}"

    # Less simple
    settings += f" --{'rect' if rect else 'no-rect'} --{'amp' if amp else 'no-amp'} --{'tskip' if tskip else 'no-tskip'}"
    settings += f" --{'b-intra' if b_intra else 'no-b-intra'} --{'weightb' if weightb else 'no-weightb'} --{'cutree' if cutree else 'no-cutree'}"
    settings += f" --rskip {int(rskip) if isinstance(rskip, bool) else rskip}"

    if isinstance(deblock, list):
        deblock = f"{str(deblock[0])}:{str(deblock[1])}"
    settings += f" --deblock={deblock}"

    # Don't need to change these lol
    settings += " --no-sao --no-sao-non-deblock --no-strong-intra-smoothing --no-open-gop"

    for k, v in kwargs.items():
        prefix = "--"
        if k.startswith("_"):
            prefix = "-"
            k = k[1:]
        settings += f" {prefix}{k.replace('_', '-')} {v}"

    settings += (" " + append.strip()) if append.strip() else ""
    return settings


def settings_builder_x264(
    preset: str = "placebo",
    crf: float = 13,
    qcomp: float = 0.7,
    psy_rd: float = 1.0,
    psy_trellis: float = 0.0,
    trellis: int = 0,
    aq_strength: float = 0.8,
    aq_mode: int = 3,
    ref: int = 16,
    bframes: int = 16,
    mbtree: bool = False,
    rc_lookahead: int = 250,
    me: str = "umh",
    subme: int = 11,
    threads: int = 6,
    merange: int = 32,
    deblock: list[int] | str = [-1, -1],
    dct_decimate: bool = False,
    append: str = "",
    **kwargs,
) -> str:
    # Simple insert values
    settings = f" --preset {preset} --crf {crf} --bframes {bframes} --ref {ref} --rc-lookahead {rc_lookahead} --me {me} --merange {merange}"
    settings += f" --aq-mode {aq_mode} --aq-strength {aq_strength} --qcomp {qcomp}"
    settings += f" --psy-rd {psy_rd}:{psy_trellis} --trellis {trellis} --subme {subme} --threads {threads}"

    # Less simple
    settings += f" {'--no-mbtree' if not mbtree else ''} {'--no-dct-decimate' if not dct_decimate else ''}"

    if isinstance(deblock, list):
        deblock = f"{str(deblock[0])}:{str(deblock[1])}"
    settings += f" --deblock={deblock}"

    for k, v in kwargs.items():
        prefix = "--"
        if k.startswith("_"):
            prefix = "-"
            k = k[1:]
        settings += f" {prefix}{k.replace('_', '-')} {v}"

    settings += (" " + append.strip()) if append.strip() else ""
    return settings


sb = settings_builder_x265
sb265 = sb
sb264 = settings_builder_x264


def file_or_default(file: PathLike, default: str, no_warn: bool = False) -> tuple[str | list[str], bool]:
    if isinstance(file, list):
        return file, False
    if os.path.isfile(file):
        file = ensure_path(file, None)
        if file.exists():
            with open(file, "r") as r:
                settings = str(r.read())
                settings = settings.replace("\n", " ")
                settings = re.sub(r"(?:-o|--output) {clip.+?}", "", settings, flags=re.I).strip()
                return settings, True

    if not no_warn:
        warn("Settings file wasn't found. Using default.", None, 3)
    return default, False


def get_props(clip: vs.VideoNode, x265: bool) -> dict[str, str]:
    crange = ColorRange.from_video(clip)
    is_limited = crange.is_limited
    c_range = crange.string if x265 else ("tv" if is_limited else "pc")
    bits = clip.format.bits_per_sample
    props = clip.get_frame(0).props
    return {
        "depth": str(bits),
        "chromaloc": str(int(ChromaLocation.from_video(clip))),
        "range": c_range,
        "transfer": Transfer.from_video(clip).string,
        "colormatrix": Matrix.from_video(clip).string,
        "primaries": Primaries.from_video(clip).string,
        "sarnum": str(props.get("_SARNum", 1)),
        "sarden": str(props.get("_SARDen", 1)),
        "keyint": str(round(clip.fps) * 10),
        "min_keyint": str(round(clip.fps)),
        "frames": str(clip.num_frames),
        "fps_num": str(clip.fps_num),
        "fps_den": str(clip.fps_den),
        "min_luma": str(16 << (bits - 8) if is_limited else 0),
        "max_luma": str(235 << (bits - 8) if is_limited else (1 << bits) - 1),
        "lookahead": str(min(clip.fps_num * 5, 250)),
    }


def fill_props(settings: str, clip: vs.VideoNode, x265: bool, sar: str | None = None) -> str:
    props = get_props(clip, x265)
    if sar is not None:
        if not isinstance(sar, str):
            sar = str(sar)
        sarnum = sar if ":" not in sar else sar.split(":")[0]
        sarden = sar if ":" not in sar else sar.split(":")[1]
    else:
        sarnum = props.get("sarnum")
        sarden = props.get("sarden")
        if sarnum != "1" or sarden != "1":
            warn(f"Are you sure your SAR ({sarnum}:{sarden}) is correct?\nAre you perhaps working on an anamorphic source?", None, 2)
    settings = re.sub(r"{chromaloc(?::.)?}", props.get("chromaloc"), settings)
    settings = re.sub(r"{primaries(?::.)?}", props.get("primaries"), settings)
    settings = re.sub(r"{bits(?::.)?}", props.get("depth"), settings)
    settings = re.sub(r"{matrix(?::.)?}", props.get("colormatrix"), settings)
    settings = re.sub(r"{range(?::.)?}", props.get("range"), settings)
    settings = re.sub(r"{transfer(?::.)?}", props.get("transfer"), settings)
    settings = re.sub(r"{frames(?::.)?}", props.get("frames"), settings)
    settings = re.sub(r"{fps_num(?::.)?}", props.get("fps_num"), settings)
    settings = re.sub(r"{fps_den(?::.)?}", props.get("fps_den"), settings)
    settings = re.sub(r"{min_keyint(?::.)?}", props.get("min_keyint"), settings)
    settings = re.sub(r"{keyint(?::.)?}", props.get("keyint"), settings)
    settings = re.sub(r"{sarnum(?::.)?}", sarnum, settings)
    settings = re.sub(r"{sarden(?::.)?}", sarden, settings)
    settings = re.sub(r"{min_luma(?::.)?}", props.get("min_luma"), settings)
    settings = re.sub(r"{max_luma(?::.)?}", props.get("max_luma"), settings)
    settings = re.sub(r"{lookahead(?::.)?}", props.get("lookahead"), settings)
    return settings


def props_args(clip: vs.VideoNode, x265: bool, sar: str | None = None) -> list[str]:
    args: list[str] = []
    props = get_props(clip, x265)
    if sar is not None:
        if not isinstance(sar, str):
            sar = str(sar)
        sarnum = sar if ":" not in sar else sar.split(":")[0]
        sarden = sar if ":" not in sar else sar.split(":")[1]
    else:
        sarnum = props.get("sarnum")
        sarden = props.get("sarden")
        if sarnum != "1" or sarden != "1":
            warn(f"Are you sure your SAR ({sarnum}:{sarden}) is correct?\nAre you perhaps working on an anamorphic source?", None, 2)

    # fmt: off
    args.extend([
        "--input-depth", props.get("depth"),
        "--output-depth", props.get("depth"),
        "--transfer", props.get("transfer"),
        "--chromaloc", props.get("chromaloc"),
        "--colormatrix", props.get("colormatrix"),
        "--range", props.get("range"),
        "--colorprim", props.get("primaries"),
        "--sar", f"{sarnum}:{sarden}"
    ])
    if x265:
        args.extend([
            "--min-luma", props.get("min_luma"),
            "--max-luma", props.get("max_luma")
        ])
    return args
    # fmt: on



from pathlib import Path
from vstools import vs
from typing import overload
from fractions import Fraction

from muxtools import do_audio as mt_audio
from muxtools import (
    PathLike,
    AudioFile,
    Encoder,
    Trimmer,
    Extractor,
    AutoEncoder,
    AutoTrimmer,
    FFMpeg,
    Trim,
    warn,
    uniquify_path,
    get_workdir,
    ensure_path,
)

from ..utils.src import src_file
from ..utils.audio import audio_async_render

__all__ = ["do_audio", "encode_audio", "export_audionode"]


def export_audionode(node: vs.AudioNode, outfile: PathLike | None = None) -> Path:
    """
    Exports an audionode to a wav/w64 file.

    :param node:            Your audionode
    :param outfile:         Custom output path if any

    :return:                Returns path
    """
    if not outfile:
        outfile = uniquify_path(Path(get_workdir(), "exported.wav"))

    outfile = ensure_path(outfile, export_audionode)
    with open(outfile, "wb") as bf:
        audio_async_render(node, bf)
    return outfile


def do_audio(
    fileIn: PathLike | src_file | vs.AudioNode,
    track: int = 0,
    trims: Trim | list[Trim] | None = None,
    fps: Fraction | PathLike | None = None,
    num_frames: int = 0,
    extractor: Extractor = FFMpeg.Extractor(),
    trimmer: Trimmer | None = AutoTrimmer(),
    encoder: Encoder | None = AutoEncoder(),
    quiet: bool = True,
    output: PathLike | None = None,
) -> AudioFile:
    """
    One-liner to handle the whole audio processing

    :param fileIn:          Input file or src_file/FileInfo or AudioNode
    :param track:           Audio track number
    :param trims:           Frame ranges to trim and/or combine, e.g. (24, -24) or [(24, 500), (700, 900)]
                            If your passed src_file has a trim it will use it. Any other trims passed here will overwrite it.

    :param fps:             FPS Fraction used for the conversion to time
                            Will be taken from input if it's a src_file and assume the usual 24 if not.
                            Also accepts a timecode (v2) file.

    :param num_frames:      Total number of frames, used for negative numbers in trims
                            Will be taken from input if it's a src_file

    :param extractor:       Tool used to extract the audio (Will default to None if an AudioNode gets passed)
    :param trimmer:         Tool used to trim the audio
                            AutoTrimmer means it will choose ffmpeg for lossy and Sox for lossless

    :param encoder:         Tool used to encode the audio
                            AutoEncoder means it won't reencode lossy and choose opus otherwise

    :param quiet:           Whether the tool output should be visible
    :param output:          Custom output file or directory, extensions will be automatically added
    :return:                AudioFile Object containing file path, delays and source
    """
    if trims is not None:
        if isinstance(fileIn, src_file):
            warn("Other trims passed will overwrite whatever your src_file has!", do_audio, 1)
        if isinstance(fileIn, vs.AudioNode):
            warn("Trims won't be applied if you pass an Audionode. Just do them yourself before this lol.", do_audio, 1)
            trims = None
            trimmer = None

    if isinstance(fileIn, vs.AudioNode):
        extractor = None
        fileIn = export_audionode(fileIn)
    elif isinstance(fileIn, src_file):
        if not trims:
            trims = fileIn.trim
        clip = fileIn.src
        num_frames = clip.num_frames
        fps = Fraction(clip.fps_num, clip.fps_den) if fps is None else fps
        fileIn = fileIn.file

    fps = Fraction(24000, 1001) if fps is None else fps
    return mt_audio(fileIn, track, trims, fps, num_frames, extractor, trimmer, encoder, quiet, output)


encode_audio = do_audio



import os
import re
import subprocess
from pathlib import Path
from muxtools import ensure_path_exists, PathLike, get_executable, run_commandline, error, info


def parse_keyframes(file: PathLike) -> list[int]:
    pattern = re.compile(r"n:(?: +)?(?P<frame>\d+).*type:(?: +)?(?P<type>.)")
    file = ensure_path_exists(file, parse_keyframes)
    frames = list[int]()
    # ffmpeg seems to be A LOT faster than ffprobe for this. (Maybe ffprobe can't multithread?)
    ffmpeg = get_executable("ffmpeg")
    out_var = "NUL" if os.name == "nt" else "/dev/null"
    args = [ffmpeg, "-hide_banner", "-i", str(file), "-map", "0:v:0", "-filter:v", "showinfo", "-f", "null", out_var]
    out = subprocess.run(args, capture_output=True, text=True)
    if out.returncode != 0:
        error("Failed to parse video keyframes using ffmpeg!", parse_keyframes)
        print(out.stderr)
        print(out.stdout)
        out.check_returncode()

    for line in out.stderr.splitlines():
        if "showinfo" in line:
            matches = re.findall(pattern, line.strip())
            if matches:
                match = matches[0]
                if match[1] == "I":
                    frames.append(int(match[0]))
    return frames


def merge_parts(last: Path, original_out: Path, keyframes: list[int], parts: list[Path], quiet: bool = True):
    # This function is partially stolen from vardautomation but with less weird abstraction
    mkvmerge = get_executable("mkvmerge")
    mkvextract = get_executable("mkvextract")

    _to_delete = set[Path]()
    mkv_parts = list[Path]()
    # Remux and split existing parts
    for kf, part in zip(keyframes, parts):
        as_mkv = part.with_suffix(".mkv").resolve()
        args = [mkvmerge, "-o", str(as_mkv), "--split", f"frames:{kf}", str(part.resolve())]
        if run_commandline(args, quiet) > 1:
            raise error("Failed to remux existing part", merge_parts)
        # mkvmerge will spit out parts with this kind of naming and there will only be two
        p_mkv001 = as_mkv.with_stem(as_mkv.stem + "-001")
        p_mkv002 = as_mkv.with_stem(as_mkv.stem + "-002")
        # we only need the first one ofc
        mkv_parts.append(p_mkv001)
        _to_delete.update([p_mkv001, p_mkv002])
    _to_delete.update(parts)

    if len(mkv_parts) < 1:
        info("Encode finished successfully without requiring other parts")
        for f in _to_delete:
            f.unlink(True)
        last.rename(original_out)
        return

    # Remux the last encoded part
    last_mkv = last.with_suffix(".mkv").resolve()
    args = [mkvmerge, "-o", str(last_mkv), str(last.resolve())]
    if run_commandline(args, quiet) > 1:
        raise error("Failed to remux last part", merge_parts)
    mkv_parts.append(last_mkv)
    _to_delete.update([last_mkv, last])

    out_mkv = original_out.with_suffix(".mkv").resolve()

    args = [mkvmerge, "-o", str(out_mkv)]
    first = True
    for part in mkv_parts:
        if not first:
            args.append("+")
        args.append(str(part.resolve()))
        first = False

    if run_commandline(args, quiet) > 1:
        raise error("Failed to remux last part", merge_parts)

    info("Extracting merged part...")
    args = [mkvextract, str(out_mkv), "tracks", f"0:{str(original_out.resolve())}"]
    if run_commandline(args, quiet) > 1:
        raise error("Failed to extract merged track!", merge_parts)

    _to_delete.add(out_mkv)

    for f in _to_delete:
        f.unlink(True)



from fractions import Fraction
from muxtools import parse_chapters_bdmv, PathLike, GlobSearch, Chapter, Chapters as Ch

from ..utils.src import src_file

__all__ = ["Chapters"]


class Chapters(Ch):
    def __init__(
        self, chapter_source: src_file | PathLike | GlobSearch | Chapter | list[Chapter], fps: Fraction | PathLike | None = None, _print: bool = True
    ) -> None:
        """
        Convenience class for chapters

        :param chapter_source:      Input either src_file/FileInfo, txt with ogm chapters, xml or (a list of) self defined chapters.
        :param fps:                 Needed for timestamp convertion. Gets the fps from the clip if src_file and otherwise assumes 24000/1001.
                                    Also accepts a timecode (v2) file.
        :param _print:              Prints chapters after parsing and after trimming.
        """
        if isinstance(chapter_source, src_file):
            clip_fps = Fraction(chapter_source.src.fps_num, chapter_source.src.fps_den)
            self.fps = fps if fps else clip_fps
            self.chapters = parse_chapters_bdmv(chapter_source.file, self.fps, chapter_source.src_cut.num_frames, _print)
            if self.chapters and chapter_source.trim:
                self.trim(chapter_source.trim[0], chapter_source.trim[1], chapter_source.src_cut.num_frames)
                if _print:
                    print("After trim:")
                    self.print()
        else:
            super().__init__(chapter_source, fps if fps else Fraction(24000, 1001), _print)



import shutil as sh
from os import PathLike
from pathlib import Path
from typing import Callable
from fractions import Fraction
from vstools import vs, core, initialize_clip, copy_signature
from muxtools import (
    Trim,
    PathLike,
    parse_m2ts_path,
    ensure_path_exists,
    warn,
    info,
    get_workdir,
    get_temp_workdir,
    clean_temp_files,
    get_absolute_track,
    TrackType,
    GlobSearch,
)


__all__ = ["src_file", "SRC_FILE", "FileInfo", "src", "frames_to_samples", "f2s"]


class src_file:
    file: PathLike
    force_lsmas: bool = False
    trim: Trim = None
    idx: Callable[[str], vs.VideoNode] | None = None
    idx_args = {}

    def __init__(
        self, file: PathLike | GlobSearch, force_lsmas: bool = False, trim: Trim = None, idx: Callable[[str], vs.VideoNode] | None = None, **kwargs
    ):
        """
        Custom `FileInfo` kind of thing for convenience

        :param file:            Either a string based filepath or a Path object
        :param force_lsmas:     Forces the use of lsmas inside of `vodesfunc.src`
        :param trim:            Can be a single trim or a sequence of trims.
        :param idx:             Indexer for the input file. Pass a function that takes a string in and returns a vs.VideoNode.
                                Defaults to `vodesfunc.src`
        """
        self.file = ensure_path_exists(file, self)
        self.force_lsmas = force_lsmas
        self.trim = trim
        self.idx = idx
        self.idx_args = kwargs

    def __index_clip(self):
        indexed = self.idx(str(self.file.resolve())) if self.idx else src(str(self.file.resolve()), self.force_lsmas, **self.idx_args)
        cut = indexed
        if self.trim:
            self.trim = list(self.trim)
            if self.trim[0] is None:
                self.trim[0] = 0
            if self.trim[1] is None or self.trim[1] == 0:
                cut = indexed[self.trim[0] :]
            else:
                cut = indexed[self.trim[0] : self.trim[1]]
            self.trim = tuple(self.trim)

        if self.file.suffix.lower() == ".dgi":
            if self.file.with_suffix(".m2ts").exists():
                self.file = self.file.with_suffix(".m2ts")
            else:
                self.file = parse_m2ts_path(self.file)

        setattr(self, "clip", indexed)
        setattr(self, "clip_cut", cut)

    @property
    def src(self) -> vs.VideoNode:
        if not hasattr(self, "clip"):
            self.__index_clip()
        return self.clip

    @property
    def src_cut(self) -> vs.VideoNode:
        if not hasattr(self, "clip_cut"):
            self.__index_clip()
        return self.clip_cut

    @copy_signature(initialize_clip)
    def init(self, *args, **kwargs) -> vs.VideoNode:
        """
        Getter that calls `vstools.initialize_clip` on the src clip for convenience

        :param kwargs:      Any other args passed to initialize_clip
        """
        return initialize_clip(self.src, *args, **kwargs)

    @copy_signature(initialize_clip)
    def init_cut(self, *args, **kwargs) -> vs.VideoNode:
        """
        Getter that calls `vstools.initialize_clip` on the src_cut clip for convenience

        :param kwargs:      Any other args passed to initialize_clip
        """
        return initialize_clip(self.src_cut, *args, **kwargs)

    def get_audio(self, track: int = 0, **kwargs) -> vs.AudioNode:
        """
        Indexes the specified audio track from the input file.
        """
        args = dict(exact=True)
        args.update(**kwargs)

        absolute = get_absolute_track(self.file, track, TrackType.AUDIO)

        return core.bs.AudioSource(str(self.file.resolve()), absolute.track_id, **args)

    def get_audio_trimmed(self, track: int = 0, **kwargs) -> vs.AudioNode:
        """
        Gets the indexed audio track with the trim specified in the src_file.
        """
        node = self.get_audio(track, **kwargs)
        if self.trim:
            if self.trim[1] is None or self.trim[1] == 0:
                node = node[f2s(self.trim[0], node, self.src) :]
            else:
                node = node[f2s(self.trim[0], node, self.src) : f2s(self.trim[1], node, self.src)]
        return node


SRC_FILE = src_file
FileInfo = src_file


def src(filePath: PathLike, force_lsmas: bool = False, delete_dgi_log: bool = True, **kwargs) -> vs.VideoNode:
    """
    Uses dgindex as Source and requires dgindexnv in path
    to generate files if they don't exist.

    :param filepath:        Path to video or dgi file
    :param force_lsmas:     Skip dgsource entirely and use lsmas
    :param delete_dgi_log:  Delete the .log files dgindexnv creates
    :return:                Video Node
    """
    filePath = ensure_path_exists(filePath, src)
    filePath = str(filePath)
    if filePath.lower().endswith(".dgi"):
        return core.lazy.dgdecodenv.DGSource(filePath, **kwargs)

    forceFallBack = sh.which("dgindexnv") is None or not hasattr(core, "dgdecodenv")

    if not force_lsmas:
        import pymediainfo as pym

        parsed = pym.MediaInfo.parse(filePath, parse_speed=0.25)
        trackmeta = parsed.video_tracks[0].to_data()
        format = trackmeta.get("format")
        bitdepth = trackmeta.get("bit_depth")
        if format is not None and bitdepth is not None:
            if str(format).strip().lower() == "avc" and int(bitdepth) > 8:
                forceFallBack = True
                warn(f"Falling back to lsmas for Hi10 ({Path(filePath).name})", src)
            elif str(format).strip().lower() == "ffv1":
                forceFallBack = True
                warn(f"Falling back to lsmas for FFV1 ({Path(filePath).name})", src)

    if force_lsmas or forceFallBack:
        return core.lsmas.LWLibavSource(filePath, **kwargs)

    path = Path(filePath)
    dgiFile = path.with_suffix(".dgi")

    if dgiFile.exists():
        return core.dgdecodenv.DGSource(dgiFile.resolve(True), **kwargs)
    else:
        info("Generating dgi file...", src)
        import os
        import subprocess as sub

        sub.Popen(f'dgindexnv -i "{path.name}" -h -o "{dgiFile.name}" -e', shell=True, stdout=sub.DEVNULL, cwd=path.parent.resolve(True)).wait()
        if path.with_suffix(".log").exists() and delete_dgi_log:
            os.remove(path.with_suffix(".log").resolve(True))
        return core.dgdecodenv.DGSource(dgiFile.resolve(True), **kwargs)


def frames_to_samples(frame: int, sample_rate: vs.AudioNode | int = 48000, fps: vs.VideoNode | Fraction = Fraction(24000, 1001)) -> int:
    """
    Converts a frame number to a sample number

    :param frame:           The frame number
    :param sample_rate:     Can be a flat number like 48000 (=48 kHz) or an AudioNode to get the sample rate from
    :param fps:             Can be a Fraction or a VideoNode to get the fps from

    :return:                The sample number
    """
    if frame == 0:
        return 0
    sample_rate = sample_rate.sample_rate if isinstance(sample_rate, vs.AudioNode) else sample_rate
    fps = Fraction(fps.fps_num, fps.fps_den) if isinstance(fps, vs.VideoNode) else fps
    return int(sample_rate * (fps.denominator / fps.numerator) * frame)


f2s = frames_to_samples


def generate_qp_file(clip: vs.VideoNode, start_frame: int = 0) -> str:
    filepath = Path(get_workdir(), f"qpfile_{start_frame}.txt")
    temp = Path(get_temp_workdir(), "qpfile.txt")
    if filepath.exists():
        info("Reusing existing QP File.")
        return str(filepath.resolve())
    info("Generating QP File...")

    clip = clip.resize.Bicubic(640, 360, format=vs.YUV410P8)
    clip = clip.wwxd.WWXD()
    if start_frame:
        clip = clip[start_frame:]
    out = ""
    for i in range(1, clip.num_frames):
        if clip.get_frame(i).props.Scenechange == 1:
            out += f"{i} I -1\n"

    with open(temp, "w") as file:
        file.write(out)

    temp.rename(filepath)
    clean_temp_files()

    return str(filepath.resolve())



from . import utils, video, extension

from .utils import *
from .video import *

from muxtools.main import *
from muxtools.audio import *
from muxtools.muxing import *
from muxtools.subtitle import *
from muxtools.utils import *

from .extension import *



import subprocess
from vstools import vs
from fractions import Fraction
from dataclasses import dataclass

from muxtools import (
    PathLike,
    ensure_path_exists,
    debug,
    error,
    get_absolute_track,
    get_executable,
    TrackType,
    frame_to_timedelta,
    VideoFile,
    VideoTrack,
    MkvTrack,
)
from muxtools import SubFile as MTSubFile
from muxtools.subtitle.sub import SubFileSelf

__all__ = ["SubFile"]


@dataclass
class SubFile(MTSubFile):
    """
    Utility class representing a subtitle file with various functions to run on.

    :param file:            Can be a string, Path object or GlobSearch.
                            If the GlobSearch returns multiple results or if a list was passed it will merge them.

    :param container_delay: Set a container delay used in the muxing process later.
    :param source:          The file this sub originates from, will be set by the constructor.
    :param encoding:        Encoding used for reading and writing the subtitle files.
    """

    def truncate_by_video(
        self: SubFileSelf, source: PathLike | VideoTrack | MkvTrack | VideoFile | vs.VideoNode, fps: Fraction | PathLike | None = None
    ) -> SubFileSelf:
        """
        Removes lines that start after the video ends and trims lines that extend past it.

        :param source:      Can be any video file or a VideoNode
        :param fps:         FPS Fraction; Will be parsed from the video by default. Also accepts a timecode (v2) file.
        """
        if isinstance(source, vs.VideoNode):
            frames = source.num_frames
            if not fps:
                fps = Fraction(source.fps_num, source.fps_den)
        else:
            if isinstance(source, VideoTrack) or isinstance(source, MkvTrack) or isinstance(source, VideoFile):
                file = ensure_path_exists(source.file, self)
            else:
                file = ensure_path_exists(source, self)
            # Unused variable, just used to have a simple validation
            track = get_absolute_track(file, 0, TrackType.VIDEO, self)
            ffprobe = get_executable("ffprobe")
            args = [ffprobe, "-v", "error", "-select_streams", "v:0", "-show_entries", "stream=r_frame_rate : stream_tags", str(file)]
            out = subprocess.run(args, capture_output=True, text=True)
            frames = 0

            for line in out.stdout.splitlines():
                if "=" not in line:
                    continue
                line = line.strip()
                if "r_frame_rate" in line and not fps:
                    fps = Fraction(line.split("=")[1])
                    debug(f"Parsed FPS from file: {fps}", self)
                elif "NUMBER_OF_FRAMES" in line:
                    line = line.split("=")[1]
                    try:
                        frames = int(line)
                        debug(f"Parsed frames from file: {frames}", self)
                    except:
                        continue

            if not fps or not frames:
                raise error(f"Could not parse frames or fps from file '{file.stem}'!", self)

        cutoff = frame_to_timedelta(frames + 1, fps)
        removed = 0
        trimmed = 0
        doc = self._read_doc()
        lines: list = doc.events
        for line in lines:
            if line.start > cutoff:
                lines.remove(line)
                removed += 1
                continue
            if line.end > cutoff:
                line.end = frame_to_timedelta(frames, fps)
                trimmed += 1

        if removed or trimmed:
            if removed:
                debug(f"Removed {removed} line{'s' if removed != 1 else ''} that started past the video", self)
            if trimmed:
                debug(f"Trimmed {trimmed} line{'s' if trimmed != 1 else ''} that extended past the video", self)
            doc.events = lines
            self.__update_doc(doc)

        return self



import shlex
import subprocess
from vstools import finalize_clip, initialize_clip, vs
from pathlib import Path
from dataclasses import dataclass
from abc import ABC, abstractmethod
from muxtools import get_executable, ensure_path_exists, VideoFile, PathLike, make_output, info, warn, get_setup_attr, ensure_path

from ..utils.types import LosslessPreset, Zone
from ..utils.src import generate_qp_file, src_file
from .resumable import merge_parts, parse_keyframes
from .settings import file_or_default, fill_props, props_args, sb264, sb265, shift_zones, zones_to_args

__all__ = ["x264", "x265", "LosslessX264", "FFV1"]


@dataclass
class VideoEncoder(ABC):
    resumable = False

    @abstractmethod
    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        """
        To actually run the encode.

        :param clip:            Your videonode
        :param outfile:         Can be a custom output file or directory.
                                The correct extension will automatically be appended.

        Returns a VideoFile object.
        If you're only interested in the path you can just do `VideoFile.file`.
        """
        ...

    def _update_progress(self, current_frame, total_frames):
        print(f"\rVapoursynth: {current_frame} / {total_frames} " f"({100 * current_frame // total_frames}%) || Encoder: ", end="")


@dataclass
class SupportsQP(VideoEncoder):
    settings: str | PathLike | None = None
    zones: Zone | list[Zone] | None = None
    qp_file: PathLike | bool | None = None
    qp_clip: src_file | vs.VideoNode | None = None
    add_props: bool | None = None
    sar: str | None = None
    quiet_merging: bool = True
    x265 = True

    def _get_qpfile(self, start_frame: int = 0) -> str:
        if not self.qp_file and not self.qp_clip:
            return ""

        if not isinstance(self.qp_file, bool) and self.qp_file is not None:
            return str(ensure_path_exists(self.qp_file, self).resolve())

        if self.qp_clip:
            if isinstance(self.qp_clip, src_file):
                self.qp_clip = self.qp_clip.src_cut
            return generate_qp_file(self.qp_clip, start_frame)

    def _init_settings(self, x265: bool):
        if not self.settings:
            s, p = file_or_default(f"{'x265' if x265 else 'x264'}_settings", sb265() if x265 else sb264())
            self.was_file = p
            self.settings = s
        else:
            s, p = file_or_default(self.settings, self.settings, True)
            self.was_file = p
            self.settings = s

        if self.add_props is None:
            self.add_props = not getattr(self, "was_file", False)

    def _update_settings(self, clip: vs.VideoNode, x265: bool):
        if self.was_file:
            self.settings = fill_props(self.settings, clip, x265, self.sar)

        self.settings = self.settings if isinstance(self.settings, list) else shlex.split(self.settings)

        if self.add_props:
            self.settings.extend(props_args(clip, x265, self.sar))

    @abstractmethod
    def _encode_clip(self, clip: vs.VideoNode, out: Path) -> Path:
        ...

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        if clip.format.bits_per_sample > (12 if self.x265 else 10):
            warn(f"This encoder does not support a bit depth over {(12 if self.x265 else 10)}.\nClip will be dithered to 10 bit.", self, 2)
            clip = finalize_clip(clip, 10)
        self._update_settings(clip, self.x265)
        out = make_output(
            Path(self.qp_clip.file).stem if isinstance(self.qp_clip, src_file) else "encoded",
            "265" if self.x265 else "264",
            "encoded" if isinstance(self.qp_clip, src_file) else "",
            outfile,
        )
        if not self.resumable:
            return VideoFile(self._encode_clip(clip, out, self._get_qpfile()))

        pattern = out.with_stem(out.stem + "_part_???")
        parts = sorted(pattern.parent.glob(pattern.name))
        info(f"Found {len(parts)} part{'s' if len(parts) != 1 else ''} for this encode")

        keyframes = list[int]()
        for i, p in enumerate(parts):
            try:
                info(f"Parsing keyframes for part {i}...")
                kf = parse_keyframes(p)[-1]
                if kf == 0:
                    del parts[-1]
                else:
                    keyframes.append(kf)
            except:
                del parts[-1]
        fout = out.with_stem(out.stem + f"_part_{len(parts):03.0f}")
        start_frame = sum(keyframes)
        info(f"Starting encode at frame {start_frame}")

        # TODO: Adjust existing zones to the new start frame

        clip = clip[start_frame:]
        self._encode_clip(clip, fout, self._get_qpfile(start_frame), start_frame)

        info("Remuxing and merging parts...")
        merge_parts(fout, out, keyframes, parts, self.quiet_merging)
        return VideoFile(out, source=self.qp_clip.file if isinstance(self.qp_clip, src_file) else None)


@dataclass
class x264(SupportsQP):
    """
    Encodes your clip to an avc/h264 file using x264.

    :param settings:            This will by default try to look for an `x264_settings` file in your cwd.
                                If it doesn't find one it will warn you and resort to the default settings_builder preset.
                                You can either pass settings as usual or a filepath here.
                                If the filepath doesn't exist it will assume you passed actual settings and pass those to the encoder.

    :param zones:               With this you can tweak settings of specific regions of the video.
                                In x264 this includes but is not limited to CRF.
                                For example (100, 300, "crf", 12) or [(100, 300, "crf", 12), (500, 750, 1.3)]
                                If the third part is not a string it will assume a bitrate multiplier (or "b")

    :param qp_file:             Here you can pass a bool to en/disable or an existing filepath for one.
    :param qp_clip:             Can either be a straight up VideoNode or a SRC_FILE/FileInfo from this package.
                                If neither a clip or a file are given it will simply skip.
                                If only a clip is given it will generate one.

    :param add_props:           This will explicitly add all props taken from the clip to the command line.
                                This will be disabled by default if you are using a file and otherwise enabled.
                                Files can have their own tokens like in vs-encode/vardautomation that will be filled in.

    :param sar:                 Here you can pass your Pixel / Sample Aspect Ratio. This will overwrite whatever is in the clip if passed.
    :param resumable:           Enable or disable resumable encodes. Very useful for people that have scripts that crash their PC (skill issue tbh)
    """

    resumable: bool = True
    x265 = False

    def __post_init__(self):
        self.executable = get_executable("x264")
        self._init_settings(self.x265)

    def _encode_clip(self, clip: vs.VideoNode, out: Path, qpfile: str, start_frame: int = 0) -> Path:
        args = [self.executable, "-o", str(out.resolve())]
        if qpfile:
            args.extend(["--qpfile", qpfile])
        if self.settings:
            args.extend(self.settings if isinstance(self.settings, list) else shlex.split(self.settings))
        if self.zones:
            if start_frame:
                self.zones = shift_zones(self.zones, start_frame)
            args.extend(zones_to_args(self.zones, False))
        args.extend(["--demuxer", "y4m", "-"])

        process = subprocess.Popen(args, stdin=subprocess.PIPE)
        clip.output(process.stdin, y4m=True, progress_update=lambda x, y: self._update_progress(x, y))
        process.communicate()
        return out


@dataclass
class x265(SupportsQP):
    """
    Encodes your clip to an hevc/h265 file using x265.

    :param settings:            This will by default try to look for an `x265_settings` file in your cwd.
                                If it doesn't find one it will warn you and resort to the default settings_builder preset.
                                You can either pass settings as usual or a filepath here.
                                If the filepath doesn't exist it will assume you passed actual settings and pass those to the encoder.

    :param zones:               With this you can tweak settings of specific regions of the video.
                                In x265 you're basically limited to a flat bitrate multiplier or force QP ("q")
                                For example (100, 300, "b", 1.2) or [(100, 300, "q", 12), (500, 750, 1.3)]
                                If the third part is not a string it will assume a bitrate multiplier (or "b")

    :param qp_file:             Here you can pass a bool to en/disable or an existing filepath for one.
    :param qp_clip:             Can either be a straight up VideoNode or a SRC_FILE/FileInfo from this package.
                                If neither a clip or a file are given it will simply skip.
                                If only a clip is given it will generate one.

    :param add_props:           This will explicitly add all props taken from the clip to the command line.
                                This will be disabled by default if you are using a file and otherwise enabled.
                                Files can have their own tokens like in vs-encode/vardautomation that will be filled in.

    :param sar:                 Here you can pass your Pixel / Sample Aspect Ratio. This will overwrite whatever is in the clip if passed.
    :param resumable:           Enable or disable resumable encodes. Very useful for people that have scripts that crash their PC (skill issue tbh)
    :param csv:                 Either a bool to enable or disable csv logging or a Filepath for said csv.
    """

    resumable: bool = True
    csv: bool | PathLike = True
    x265 = True

    def __post_init__(self):
        self.executable = get_executable("x265")
        self._init_settings(self.x265)

    def _encode_clip(self, clip: vs.VideoNode, out: Path, qpfile: str, start_frame: int = 0) -> Path:
        args = [self.executable, "-o", str(out.resolve())]
        if self.csv:
            if isinstance(self.csv, bool):
                show_name = get_setup_attr("show_name", "")
                csv_file = Path(show_name + f"{'_' if show_name else ''}log_x265.csv").resolve()
            else:
                csv_file = ensure_path(self.csv, self)
            args.extend(["--csv", str(csv_file)])
        if qpfile:
            args.extend(["--qpfile", qpfile])
        if self.settings:
            args.extend(self.settings if isinstance(self.settings, list) else shlex.split(self.settings))
        if self.zones:
            if start_frame:
                self.zones = shift_zones(self.zones, start_frame)
            args.extend(zones_to_args(self.zones, True))
        args.extend(["--y4m", "-"])

        process = subprocess.Popen(args, stdin=subprocess.PIPE)
        clip.output(process.stdin, y4m=True, progress_update=lambda x, y: self._update_progress(x, y))
        process.communicate()
        return out


@dataclass
class LosslessX264(VideoEncoder):
    """
    Uses x264 to encode clip to a lossless avc stream.

    :param preset:          Can either be a string of some x264 preset or any of the 3 predefined presets.
    :param settings:        Any other settings you might want to pass. Entirely optional.
    :param add_props:       This will explicitly add all props taken from the clip to the command line.
    """

    preset: str | LosslessPreset = LosslessPreset.MIDDLEGROUND
    settings: str | None = None
    add_props: bool = True

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        out = make_output("lossless", "264", user_passed=outfile)
        match self.preset:
            case LosslessPreset.SPEED:
                preset = "ultrafast"
            case LosslessPreset.COMPRESSION:
                preset = "veryslow"
            case LosslessPreset.MIDDLEGROUND:
                preset = "medium"
            case _:
                preset = self.preset
        settings = ["--preset", preset, "--qp", "0"]
        if clip.format.bits_per_sample > 10:
            warn(f"This encoder does not support a bit depth over 10.\nClip will be dithered to 10 bit.", self, 2)
            clip = finalize_clip(clip, 10)

        if self.settings:
            settings.extend(shlex.split(self.settings))
        avc = x264(settings, add_props=self.add_props, resumable=False)
        avc._update_settings(clip, False)
        avc._encode_clip(clip, out, None, 0)
        return VideoFile(out)


@dataclass
class FFV1(VideoEncoder):
    """
    Uses ffmpeg to encode clip to a lossless ffv1 stream.

    :param settings:        Can either be a string of your own settings or any of the 3 presets.
    :param ensure_props:    Calls initialize_clip on the clip to have at the very least guessed props
    """

    settings: str | LosslessPreset = LosslessPreset.MIDDLEGROUND
    ensure_props: bool = True

    def __post_init__(self):
        self.executable = get_executable("ffmpeg")

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        bits = clip.format.bits_per_sample
        if bits > 10:
            warn(f"You are encoding FFV1 with {bits} bits. It will be massive.", self, 1)
        if self.ensure_props:
            clip = initialize_clip(clip, bits)
            clip = finalize_clip(clip, bits)
        _base = "-coder 1 -context 0 -g 1 -level 3 -threads 0"
        match self.settings:
            case LosslessPreset.SPEED:
                self.settings = _base + " -slices 30 -slicecrc 0"
            case LosslessPreset.COMPRESSION:
                self.settings = _base + " -slices 16 -slicecrc 1"
            case LosslessPreset.MIDDLEGROUND:
                self.settings = _base + " -slices 24 -slicecrc 1"
            case _:
                self.settings = self.settings

        out = make_output("encoded_ffv1", "mkv", user_passed=outfile)

        args = [self.executable, "-f", "yuv4mpegpipe", "-i", "-", "-c:v", "ffv1"]
        if self.settings:
            args.extend(shlex.split(self.settings))
        args.append(str(out))

        process = subprocess.Popen(args, stdin=subprocess.PIPE)
        clip.output(process.stdin, y4m=True, progress_update=lambda x, y: self._update_progress(x, y))
        process.communicate()
        return VideoFile(out)



from pathlib import Path
import re
from vstools import vs
from shlex import split
from numpy import arange

from .encoders import SupportsQP, VideoEncoder, x265
from ..utils.src import SRC_FILE, src
from muxtools import make_output

__all__ = ["SettingsTester"]


def settings_to_dict(settings: str | list[str]) -> dict[str, str | None]:
    args = split(settings) if isinstance(settings, str) else settings
    args_dict = dict[str, str | None]()

    for i, arg in enumerate(args):
        if not arg.startswith("-") or arg[1:].isnumeric():
            continue
        nextarg = None if len(args) < i + 2 else args[i + 1]
        if not nextarg or (nextarg.startswith("-") and not nextarg[1:].isnumeric()):
            args_dict.update({arg: None})
        else:
            args_dict.update({arg: nextarg})

    return args_dict


def resolve_var(var: str) -> list[str]:
    has_equals = False
    if "=" in var:
        var = var.split("=")[1]
        has_equals = True
    var = var[1:-1]

    if "/" in var:
        start = float(var.split("/")[0].strip())
        end = float(var.split("/")[1].strip())
        step = var.split("/")[2].strip()
        decimals = len(step.split(".")[1]) if "." in step else None
        step = float(step)
        return [("=" if has_equals else " ") + str(round(val, decimals)) for val in arange(start, end + step, step) if round(val, decimals) <= end]
    else:
        return [f"{'=' if has_equals else ' '}{val.strip()}" for val in var.split(",")]


def generate_settings(settings: str) -> list[tuple[str, str]]:
    yes = ["y", "yes", "true"]
    no = ["n", "no", "false"]
    pattern = re.compile(r"\[(.*?(?:\/|\,).*?)\]")
    variable_args = dict[str, str]({k: v for (k, v) in settings_to_dict(settings).items() if (v and pattern.search(v)) or pattern.search(k)})
    args_dict = dict[str, str | None]({k: v for (k, v) in settings_to_dict(settings).items() if k not in variable_args.keys()})

    settings_list = list[tuple[str, str]]()

    for k, v in variable_args.items():
        sett = settings_list.copy() if settings_list else [("", " ".join([f"{k} {v if v else ''}".strip() for (k, v) in args_dict.items()]))]
        settings_list.clear()
        resolved = resolve_var(v if v else k)
        if "=" in k:
            k = k.split("=")[0]
        for s in sett:
            for res in resolved:
                if res.strip().lower() in yes + no:
                    if res.strip().lower() in yes:
                        key = k if "--no-" not in k else f"--{k[5:]}"
                    else:
                        key = k if "--no-" in k else f"--no-{k[2:]}"
                    key_name = key[2:] if key.startswith("--") else key[1:]
                    settings_list.append((s[0] + f" {key_name}", f"{s[1]} {key}"))
                else:
                    key_name = k[2:] if k.startswith("--") else k[1:]
                    name = s[0] + f" {key_name}{'=' if '=' not in res else ''}{res.strip()}"
                    settings_list.append((name.strip(), f"{s[1]} {k}{res}"))

    return settings_list


class SettingsTester:
    """
    A utility class for making test encodes.
    This can automatically parse strings like

    `--preset veryfast --crf [14/15/0.5]`
    and will then run encodes with CRF 14, 14.5 and 15.


    The same works with non stepping options like

    `--preset [fast,veryfast,slow,slower]`
    or
    `--sao [true,false]` (can also use yes/no and y/n for these)


    Keep in mind that this will create an encode for EVERY combination and not some order.

    `--preset [fast,veryfast,ultrafast] --crf [14/15/0.5]`
    For example will end up with 9 encodes.
    """

    encoder: VideoEncoder
    encodes = list[tuple[str | None, str]]()
    qp_file: str | None = None

    def __init__(self, settings: str | list[str], encoder: VideoEncoder | None = None, qp_clip: SRC_FILE | vs.VideoNode | None = None) -> None:
        if not encoder:
            self.encoder = x265("--kek")
        else:
            self.encoder = encoder

        if isinstance(self.encoder, SupportsQP) and qp_clip:
            self.encoder.qp_clip = qp_clip
            self.qp_file = self.encoder._get_qpfile()

        if isinstance(settings, str):
            self.encodes = generate_settings(settings)
        else:
            self.encodes = [(None, s) for s in settings]

    def run(self, clip: vs.VideoNode, output_clips: bool = True) -> None:
        """
        Runs all encodes with the settings specified/generated.

        :param output_clips:        Will index and output clips with proper naming if vspreview is installed.
                                    This might obviously end up using quite a lot of ram.
        """
        for encode in self.encodes:
            encoder = self.encoder.__class__(settings=encode[1])
            encoder.resumable = False
            if isinstance(self.encoder, SupportsQP) and self.qp_file:
                encoder.qp_file = self.qp_file

            out = make_output("encode", "test", suffix="" if not encode[0] else f"[{encode[0]}]")

            f = encoder.encode(clip, out)
            if output_clips:
                done = src(f.file, True)
                try:
                    from vspreview import set_output

                    set_output(done, name=encode[0], cache=False)
                except:
                    done.set_output(len(vs.get_outputs()))
